/*++++++++++++++
.IDENTIFICATION obj@.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    SIMBAD
.KEYWORDS       
.VERSION  1.0   15-May-1997
.VERSION  1.1   24-Apr-1998: Convert \objS{SIMBAD Name}{Name in text}
.VERSION  1.2   13-Oct-1998: Take care of \objName \objS{Name #1}{#1}
		==> A voir: \bibobj
		==> A voir: \object  uses the %R
.VERSION  1.3   13-Mar-1999: \objA dans le texte des abstracts...
.VERSION  1.4   11-Jun-1999; 3 parameters in \objac{address}{name}{J2000}
.VERSION  1.5   01-Mar-2000: Trace the bibcode
.VERSION  1.6   10-Mar-2000: \objM{  ,  , }{}  --- LISTE
			(unless option -2)
.VERSION  1.7   09-Nov-2001: Accept , or ; in objS (no need for \objM)
.VERSION  1.8   15-Mar-2002: Trap the SIGPIPE Signal
.VERSION  1.9   20-Sep-2005: Accept notes instead of bibcodes
.VERSION  1.91  15-Nov-2006: -version
.VERSION  1.92  26-Feb-2007: option -m (list of "multiple" names)
.VERSION  1.93  28-Jun-2007: Cosmetics (-Wall)
.VERSION  1.94  01-Dec-2011: \objP = object with Prefix
.VERSION  1.95  05-Mar-2012: getex: don't accept a paragraph (\n\n)
.COMMENTS       Transform \oject{name}  into \obja{address}{name}{
---------------*/

#define VERSION "1.95"

#include <time.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <simcli.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>	/* OPEN */
#include <sys/stat.h>
#include <fcntl.h>
#ifndef O_BINARY
#define O_BINARY 0
#endif
#define HOSTNAME	"130.79.128.4"
#define SERVICE		"1674"

#define MAXARGLEN	(BUFSIZ-4)

static char *macros[] = {
   "\\object1" ,
   "\\objS2",
   "\\tagobj1" ,
   "\\tagobjS2",
   "\\objName1+Name ",
   "\\objP2.",				/* \objP{#1}{#2} = \objS{#1 #2}{#2} */
   "\\objM2,",				/* List */
   (char *)0 
} ;
static char *texarg[10] ;		/* #0 = Macro, ...   */
static char seplist = 0 ;		/* set by getex list */
static int  hs = -1;			/* Handle for Simbad */
static char opt2 = 0 ;			/* -2 option  ...... */

static FILE *logfile ;			/* Contents of SIMBAD names */

char *getenv(), *getpass();
double atof() ;

static char *hostname, *service, *username, *password;

static char usage[] = "\
Usage: obja [-2] [-log file] [-m file] [-Hhost] [-Sport] [-Uuser] [-Ppasswd] \
...\n\
" ;
static char help[] = "\
   -2: to limit \\obja to 2 arguments \n\
       (default 3 args \\obja{@}{name}{J2000-pos})\n\
 -log: log in an auxiliary file the objects queried\n\
   -m: file containing a list of known multiple objects (comment char is !)\n\
   -H specifies the hostname to connect to (default: simbad)\n\
   -S specifies the service  to connect to (default: 1674)\n\
   -U specifies the Simbad username (default prompted)\n\
   -P specifies the Simbad password (default prompted)\n\
" ;

static char **aargv ; 
static int   aargc ;
static char *multiples;
static char **mules;		/* Sorted multiple objects in SIMBAD */
static int  nules;		/* Number of elements in mules */

typedef int (*SORT_FCT)(/* const void *, const void * */);

/*============================================================================
 *		Help
 *============================================================================*/

static char *Serror(int handle)
/*++++++++++++++++
.PURPOSE  Interface to simbad_error
.RETURNS  A valid error !
-----------------*/
{
  char *p ;
    p = simbad_error(handle) ;
    return(p ? p : "(?!?!? Undefined SIMBAD Error ?!?!?)") ;
}

/*============================================================================
 *		General Utility Routines
 *============================================================================*/

static int strloc(str, c)
/*+++++++
.PURPOSE Locate the first occurence of character `c'
.RETURNS Index of `c' in str; length of str if `c' not found.
---------*/
  char *str;	/* IN: string to scan	*/
  char c;	/* IN: char to locate	*/
{
  register char *p;

    for (p=str; *p; p++)
  	if (*p == c)	break;
    return(p-str);
}

#if 0	/* Exists */
static char *strdup(str)
/*++++++++++++++++
.PURPOSE  Clone a string
.RETURNS  Newly allocated string
-----------------*/
  char *str ;
{
  char *newstr ;
    newstr = (char *) malloc(strlen(str)+1) ;
    strcpy(newstr,str) ;
    return(newstr) ;
}
#endif

static char *sexa(double v, char *opt)
/*++++++++++++++++
.PURPOSE  Edit a number in sexagesimal, opt = "[+]nd"
.RETURNS  Edited number
-----------------*/
{
  static double roundup[] = { 5.e-1, 5.e-2, 5.e-3, 
      5.e-4, 5.e-5, 5.e-6, 5.e-7, 5.e-8, 5.e-9, 5.e-10 };
  static char buf[30] ;
  char sign ;
  int nd, i; 
  char *p ;
    sign = 0 ; p = buf ;
    if (v < 0) { v = -v ; sign = '-'; }
    if (*opt == '+') { opt++ ; if (sign == 0) sign = '+'; }
    if (sign) *(p++) = sign ;
    nd = isdigit(*opt) ? *opt - '0' : 0 ;	/* Decimals */
    v += roundup[nd]/3600. ;

    i = v ; v -= i ; v *= 60. ;
    *(p++) = '0' + (i/10) ;
    *(p++) = '0' + (i%10) ;
    *(p++) = ' ' ;
    i = v ; v -= i ; v *= 60. ;
    *(p++) = '0' + (i/10) ;
    *(p++) = '0' + (i%10) ;
    *(p++) = ' ' ;
    i = v ; v -= i ; 
    *(p++) = '0' + (i/10) ;
    *(p++) = '0' + (i%10) ;
    if (nd) {
        *(p++) = '.' ;
	while (--nd >= 0) {
	    v *= 10. ;
	    i = v ; v -= i ; 
	    *(p++) = '0' + i ;
	}
    }
    *p = 0 ;
    return(buf) ;
}

static char *load_file(char *name)
/*++++++++++++++++
.PURPOSE  Load a file completely into memory
.RETURNS  The loaded text terminated by NULL / NULL if error
-----------------*/
{
  int fh;
  int count; 
  char *buffer;

    fh = open(name, O_BINARY);
    if (fh<0) {
	fprintf(stderr, "****load_file(%s)", name); 
	perror(""); 
	return ((char *)0);
    }
    count = lseek(fh, 0L, 2); 
    lseek(fh, 0L, 0);

    if ((buffer = malloc(count+1))) {
	if (read(fh, buffer, count) != count) {
	    fprintf(stderr, "****load_file(%s[%dbytes])", name, count);
	    perror("");
	    free(buffer);
	    buffer = (char *)0;
	}
	buffer[count] = 0;
    }
    close(fh);

    return(buffer);
}

static char **array(char *list, int sep, int comment_char, int *count)
/*++++++++++++++++
.PURPOSE  Transform a list (separator 'sep') into an array
.RETURNS  An array corresponding to the list
-----------------*/
{
  int n=0, len, i;
  char *p, **a, **vec;
    
    /* Count the number of valid lines */
    for (p=list; *p; p+=len) {
	len = strloc(p, sep);
	if ((len>0) && (*p != comment_char)) n++;
	if (p[len]) p++;
    }

    /* Allocate the array */
    a = vec = (char **)malloc((n+1)*sizeof(char *));

    /* Fill the array */
    for (p=list; *p; p+=len) {
	len = strloc(p, sep);
	if (p[len]) p[len++] = 0;	/* Replace separator */
	if (*p == 0) continue;		/* Empty line */
	i = strloc(p, comment_char);	/* Comment... */
	p[i--] = 0;
	while ((i>=0) && isspace(p[i])) p[i--] = 0;
	if (i>0) *(a++) = p;
    }

    len = (a-vec);			/* Number of elements */
    *a = (char *)0;			/* Terminate array    */
    if (count) *count = len;		/* Number of elements */

    return(vec);
}

static int acmp(char **a, char **b)
/*++++++++++++++++
.PURPOSE  Compare 2 strings
.RETURNS  Difference
.REMARKS  Use sort / bsearch
-----------------*/
{
    return(strcmp(*a, *b));
}

static int simquery(char *question)
/*++++++++++++++++
.PURPOSE  Interface to simbad_query
.RETURNS  See simbad_query
.REMARKS  Log
-----------------*/
{
    if (logfile) fprintf(logfile, "%s\n", question), fflush(logfile) ;
    return(simbad_query(hs, question, (char *)0));
}

/*============================================================================
 *		Extract arguments from TeX Macro
 *============================================================================*/
static int getex(macro, in)
/*++++++++++++++++
.PURPOSE  Get a macro, extract the arguments
.RETURNS  The number of arguments / -1 = not a known macro
-----------------*/
  char *macro ;		/* IN: Macro to find out, starts by \ */
  FILE *in ;		/* IN: File to read for args */
{
  char **a, *m, *p, *p1, *e ;
  int len, narg, iarg, c, i ;

    len = strlen(macro) ;
    narg = -1 ;
    for (a=macros; *a; a++) {
	if (strncmp(*a, macro, len) != 0) continue ;
	if (isalpha((*a)[len])) continue ;
	m = *a + len ;	/* Number of arguments */
	narg = atoi(m) ;
	while (isdigit(*m)) m++ ;
	break ;
    }
    if (! *a) return(-1) ;
    texarg[0] = *a ;
    iarg = 0;

    seplist = 0 ;
    if (*m == ',') 		/* a LIST */
	seplist = *(m++) ;

    for (i=1; i<=narg; i++) {
	p1 = (char *)0 ;
	if ((*m == '.') && (i>1)) { 	/* V1.94: Concatenation */
	    if (p<e) *(p++) = ' ';
	    if (i == narg) {		/* Add the truncated name as last arg */
		p1 = p;
		m++;
	    }
	}
	else {
	    ++iarg;
	    if (!texarg[iarg]) texarg[iarg] = malloc(MAXARGLEN+1) ;
	    p = texarg[iarg] ; e = texarg[iarg] + MAXARGLEN ;
	}
	if (*m == '+') {	/* 2nd arg = first */
	   for (++m; *m && (p<e); p++, m++) *p = *m ;
	   p1 = p ;
	}
	c = getc(in) ;  len = -1 ;
	if (c != '{' /*'}'*/) {
	    *(p++) = c ;
	}
	else for (len=1 ; (len>0) && (p<e) ; p++) {
	    *p = c = getc(in) ;
	    if (c == '{') len++ ;
	    else if (c == '}') len-- ;
	    /* Added V1.95: don't accept \n\n */
	    else if (c == '\n') { 
	      static char err[] = "***Missing.Brace?***"; int j;
		c = getc(in); ungetc(c, in);
		if (c != '\n') continue; 
		fprintf(stderr, "#+++end-of-paragraph in %s{", macro);
		fwrite(texarg[iarg], 1, (p-texarg[iarg]), stderr);
		fprintf(stderr, "...}\n");
		for (j=0; err[j] && (p<e); j++) *(p++) = err[j];
		break;
	    }
	    else if (c == EOF) break ;
	}
	if (len == 0) {		/* Remove closing brace */
	    --p; 
	}
	if (len > 0) ungetc(c, in);
	*p = 0 ;

	if (p1)	 {		/* Duplication of Argument */
	    ++iarg;
	    if (!texarg[iarg]) texarg[iarg] = malloc(MAXARGLEN+1) ;
	    strcpy (texarg[iarg], p1) ;
	    p1 = (char *)0 ;
	}
    }
    if ((p = texarg[++iarg])) *p = 0; /*texarg[i] = (char *)0 ;*/

    /* VERSION Nev. 2001: Accept , when several Objetrs */
    if ((seplist == 0) && (narg == 2)) for 
	(p = texarg[1]; (*p) && (!seplist); p++) {
	if ((*p == ',') || (*p == ';')) seplist = *p ;
    }
    return(iarg) ;
}

/*============================================================================
 *		Start up SIMBAD
 *============================================================================*/

static int init()
/*++++++++++++++++
.PURPOSE  Check hostname, etc, and open the Simbad session
.RETURNS  
.REMARKS  
-----------------*/
{
  int hs;	/* Handle for Simbad */

    if (!hostname) hostname = HOSTNAME;
    if (!service)  service = SERVICE;
#if 0	/* Old version: ask for username/password */
    if (!username) {
	username = getenv("USER");
    	fprintf(stderr,"simbad client userid [%s]: ", username) ;
	if (!fgets(buf, sizeof(buf), stdin)) execv(aargc, aargv) ;
	if (*buf) username = buf;
    }
    if (!password) 
	password = getpass("simbad client passwd: ") ;
#else	/* Take defaults */
    if (!username) username = "v1332viz";
    if (!password) password = "sim4Viz";
#endif
    hs = simbad_connect(hostname, service, username, password);
    if (hs < 0) {
	fprintf(stderr, "****Error Connecting Simbad: %s\n", Serror(0));
	exit(1); 	/* execv(aargc, aargv) */
    }
    return(hs);
}

/*============================================================================
 *		Filter the input
 *============================================================================*/

static void prm(char *macro)
/*++++++++++++++++
.PURPOSE  Print the original macro
.RETURNS  ---
-----------------*/
{
  char **a ;
    printf("%s", macro) ;
    for (a = texarg+1; *a; a++) printf("{%s}", *a) ;
}

static char *querySimbad; 

static void onSimbadDeath(int sig)
/*++++++++++++++++
.PURPOSE  In case of SGIPIPE
.RETURNS  ---
-----------------*/
{
    perror("****onSimbadDeath") ;
    fprintf(stderr, "****onSimbadDeath(%d), querySimbad='%s'\n", sig,
	querySimbad ? querySimbad : "(nil)") ;
    exit(1) ;
}

static void filter(FILE *in)
/*++++++++++++++++
.PURPOSE  Filter the file
.RETURNS  ---
-----------------*/
{
  static char bibcode[64] ;
  char macro[60], *macrout, bsim[MAXARGLEN+6];
  char *p, *q, *oname, *sname, *snext=0, *pj, *errtxt, section, sep1=0;
  time_t tim; 
  int c, i, found, colno ;
  char with_coo[3];

    /* Connect to SIMBAD and Print the Prologue	*/
    if (hs < 0) {
	hs = init();
    	time(&tim);
    	printf("#Filtered by 'obja'(%s)\t\t%s", hostname, ctime(&tim));
    	if (hs < 0) printf("#****Can't connect to SIMBAD\n") ;
	else signal(SIGPIPE, onSimbadDeath) ;
    }
	    
    section = 0 ;
    querySimbad = bsim ;
    strcpy(bibcode, "(no bibcode yet)") ;

    colno = 0 ;		/* Indicates beginning of line */
    while ((c = getc(in)) != EOF) {
	/* Try to find out in which section we are... */
	if (c == '\n') {	
	    putchar(c) ;
	    colno =0 ;
	    continue ;
	}
	colno++ ;
	if ((colno == 1) && (c == '%')) {	/* Which section ? */
	    putchar(c) ;
	    section = c = getc(in) ;
	    putchar(c) ;
	    /* Get the BibCode (or a reference in V1.9) */
	    if (section != 'R') continue ;
	    c = getc(in) ;
	    while (isspace(c)) putchar(c), c = getc(in) ;
	    for (i=0; isgraph(c) && (i < sizeof(bibcode)-1); i++) {
		    putchar(c) ; bibcode[i] = c ;
		    c = getc(in) ;
	    }
	    putchar(c) ; bibcode[i] = 0 ;
	    continue ;
	}
	if (c != '\\') { putchar(c); continue ; }
	if (hs < 0) { putchar(c); continue ; }
	p = macro ; *(p++) = c ;
	c = getc(in) ;
	while(isalpha(c)) {
	    *(p++) = c ;
	    c = getc(in) ;
	}
	ungetc(c, in);
	*p = 0 ;
	i = getex(macro, in) ;
	if (i <= 0) { printf("%s", macro) ; continue ; }

	/* I've matched the arguments: 1=SIMBAD name, 2=Original name 
	   But for LISTs, 1=List of SIMBAD Names, 2=Global Name
	*/
	oname = i >= 2 ? texarg[2] : texarg[1] ;
	sname = texarg[1] ;			/* SIMBAD Name */

	if (seplist) {	/* LIST: define the s / onames */
	    printf("%s (", oname) ;
	    snext = sname + strloc(sname, seplist) ;
	    sep1 = *snext ; *snext = 0 ;
	    while (*sname == ' ') putchar(' '), sname++ ;
	    oname = sname ;
	}

	/* When an object is found, the ouput macro may be "\objac" or
	   "\tagobjac" depending on the input macro
	*/
	macrout = macro[1] == 't' ? "\\tagobj" : "\\obj" ;

      Query1Name:
	/* Remove the { } for Simbad Query */
	strcpy(bsim, sname) ;
	for (p=q=bsim; *p; p++) {
	    if (*p == '{') continue ;
	    if (*p == '}') continue ;
	    *(q++) = *p ;
	}
	*q = 0 ;

	/* Send the Question to Simbad */
	found = simquery(bsim) ;

	if ((found < 0) && (errtxt = simbad_error(hs))) {
						/* Could be a crash ? */
	    if (strstr(errtxt, "crash")) {
	        fprintf(stderr, "****%s{%s}[%s] %s\n", macro, sname, 
		  bibcode, errtxt) ;
		simbad_close(hs) ;
		fprintf(stderr, "****Trying a reconnection....") ;
		hs = init();
		fprintf(stderr, " newhandle=%d\n", hs) ;
		if (logfile) fprintf(logfile, "####Reconnected ==> handle=%d\n",
		    hs);
	        if (hs < 0) {
		    fprintf(stderr, "****FAILED to RECONNECT****\n") ;
		    exit(1) ;
	        }
	        /* Query Again ... */
	        errtxt = (char *)0 ;
	        found = simquery(bsim) ;
	        if (found < 0) errtxt = simbad_error(hs) ;
	        if (errtxt) {
	            fprintf(stderr, 
		      "****%s{%s}[%s] %s **AGAIN**\n****I GIVE UP!\n",
		      macro, sname, bibcode, errtxt) ;
		    exit(1) ;
		}
	    }
	}
	if (found <= 0)  {	/* Try with p. */
	    for (p = bsim+strlen(bsim); p >= bsim; p--) p[3] = p[0] ;
	    bsim[0] = 'p';
	    bsim[1] = '.' ;
	    bsim[2] = ' ' ;
	    found = simquery(bsim) ;
	    if (found < 0) {	/* Try Name	*/
	        for (p = bsim+strlen(bsim); p >= bsim; p--) p[2] = p[0] ;
		bsim[0] = 'N' ;
		bsim[1] = 'a' ;
		bsim[2] = 'm' ;
		bsim[3] = 'e' ;
		bsim[4] = ' ' ;
	        found = simquery(bsim) ;
	    }
	}
	if (found < 0) {
	    fprintf(stderr, "****%s{%s}[%s]%s\n", macro, sname, 
		bibcode, Serror(hs)) ;
	    printf("%snS{%s}", macrout, oname);	/* NOT in SIMBAD */
	    continue;
	}

	if (found == 0) {		/* \objnS or \tagobjnS   */
	    fprintf(stderr, "++++%snS{%s}\n", macrout, oname) ;
	    printf("%snS{%s}", macrout, oname);	/* NOT in SIMBAD */
	    continue ;
	}

	if (found > 1) { char **astr;
	    /* V1.92: Verify multiplicity is OK */
	    astr = &sname;
	    if (mules && bsearch(astr, mules, nules, sizeof(char *), 
			(SORT_FCT)acmp)) ;
	    else fprintf(stderr, 
	        "++++%s{%s}[%s] gives %d SIMBAD objects\n",
	        macro, sname, bibcode, found) ;
	}

	if (simbad_retrieve(hs, 1) < 0) {
	    fprintf(stderr, "****%s{%s}%s: %s\n", macro, sname, Serror(hs),
		bibcode);
	    prm(macro) ;
	    continue;
	}

	/* Are there coordinates? If yes, with_coo = 'a' ; 'A' otherwise */
	with_coo[0] = 'a' ;
	with_coo[1] = 'c' ;
	with_coo[2] = 0 ;
	if (simbad_findata(hs, "J", "") <= 0)
	    with_coo[0] = 'A', with_coo[1] = 0, pj = (char *)0 ;
	else pj = simbad_getdata(hs, 0);

	/* Outside the Keywords section, no Aladin image ... */
	if (opt2) {	/* 2 arguements only */
	    with_coo[1] = 0 ;
	    if (section != 'K') with_coo[0] = 'A' ;
	}

	if (simbad_findata(hs, "@", "") < 0) {
	    fprintf(stderr, "****%s{%s}%s: %s\n", macro, sname, Serror(hs),
		bibcode);
	    prm(macro) ;
	    continue;
	}
	p = simbad_getdata(hs, 0);
	if (!p) {
	    fprintf(stderr, "****%s{%s}%s: %s\n", macro, sname, Serror(hs),
		bibcode);
	    prm(macro) ;
	    continue;
	}
	while (*p == ' ') p++ ;
	printf("%s%s{%s}{%s}", macrout, with_coo, p, oname) ;
	if (with_coo[0] == 'a') {	/* Edit sexagesimal Position */
	    while (isspace(*pj)) pj++ ;
	    printf("{%s", sexa(atof(pj)/15., "4")) ;
	    while (isgraph(*pj)) pj++ ; 
	    while (isspace(*pj)) pj++ ;
	    printf(" %s}", sexa(atof(pj), "+3")) ;
	}

	/** LIST: Look for next name */
	if (seplist) {
	    if ((*snext = sep1)) 		/* Find next item in list */
		for (++snext; *snext == ' '; snext++) ;
	    if (*snext) {		/* YES, there is an item! */
		sname = snext ;
		putchar(seplist) ;
	        snext = sname + strloc(sname, seplist) ;
	        sep1 = *snext ; *snext = 0 ;
	        while (*sname == ' ') putchar(' '), sname++ ;
	        oname = sname ;
		goto Query1Name ;
	    }
	    else putchar(')') ;
	}
    }
    if (ferror(in)) perror("****Error obja in filter input file") ;
}

/*===========================================================================*/
int main(int argc, char **argv) 
/*++++++++++++++++
.PURPOSE  
.RETURNS  to OS
.REMARKS  Options -Sservice -Uuser -Ppasswd
-----------------*/
{
  FILE *file ;
  char *parm;
  int processed_files = 0 ;

    /* Check the Arguments */
    aargc = argc ; aargv = argv ;
    while (--argc > 0) {
	parm = *++argv;
	if (*parm == '-') switch(parm[1]) {
	    case 'h':	/* Help */
		printf("%s%s", usage, help);
		exit(0);
	    case 'v':
		printf("obja (CDS) Version %s\n", VERSION);
		exit(0);
	    case 'l':	--argc, ++argv;
		logfile = fopen(*argv, "w") ;
		if (!logfile) {
		    fprintf(stderr, "****(-log) ");
		    perror(*argv) ;
		    exit(1) ;
		}
		continue ;
	    case 'm':   --argc, ++argv;
		if (multiples) { free(multiples); free(mules); } 
		multiples = load_file(*argv);
		if (!multiples) exit(1);
		mules = array(multiples, '\n', '!', &nules);
		qsort(mules, nules, sizeof(char *), (SORT_FCT)acmp);
		continue;
	    case '2':	opt2 = 1 ; continue ;
	    case 'S':	service  = parm+2; continue;
	    case 'H':	hostname = parm+2; continue;
	    case 'U':	username = parm+2; continue;
	    case 'P':	password = strdup(parm+2); 
			for (parm += 2; *parm; parm++) *parm = 0;
			continue;
	    default:
		fprintf(stderr, "****Unknown option: %s\n%s", parm, usage);
		
	}
	file = fopen(parm, "r") ;
	if (!file) { fprintf(stderr, "****"); perror(parm) ; exit(1) ; }
	filter(file) ;
	processed_files++;
	continue;
    }
    if (!processed_files) filter(stdin) ;
    simbad_disconnect(hs);
    exit(0);
}
