/*++++++++++++++
.IDENTIFICATION rsimbad.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    SIMBAD
.KEYWORDS
.VERSION  1.0   25-Nov-1994
.VERSION  1.1   03-Jan-1996: -1 option.
.VERSION  1.2   09-Feb-2003:  Accept any limitation number -- -p option
.VERSION  1.3   31-May-2006: -n option
.VERSION  1.4   15-Sep-2007: -q option
.VERSION  1.5   30-Oct-2007: -i option
.VERSION  1.6   21-Aug-2008: accept '*' in astrotypes
.VERSION  1.7   16-Dec-2009: Use readline + 64bits
.VERSION  1.71  20-Feb-2010: fflush was missing
.COMMENTS       Edit results from Simbad   ==> SERVER   1674
---------------*/

#define VERSION "1.71"

#include <time.h>
#include <math.h>		/* System	*/
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <simcli.h>
#include <string.h>
#include <unistd.h>
#include <strings.h>
#define HOSTNAME	"130.79.128.4"
#define SERVICE		"1674"

#define MIN(a,b)		((a) <= (b) ? (a) : (b))
#define ITEMS(a)	(sizeof(a)/sizeof(a[0]))
#ifndef HAS_READLINE
#define HAS_READLINE 0
#endif
#if HAS_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#define HISTORY_FILE ".queries"
char *history_file;
#endif

char *getenv(), *getpass();

static char *hostname, *service, *username, *password;

	/* The list of parameters contain explanations;
	   if explanation starts with '*', is just sent to Simbad
	   if explanation starts with ',', convert ',' into blanks
	*/
static char *params[] = {		/* Allowed parameters */
    "astrotypes",  ",(comma)-separated desired astrotypes\n\
                      (# for number of values, wild cards ? * [] acceptable)",
    "cooradius",	"*Selection radius (degrees)",
    "cooequinox",	"*Equinox of positions (default = J2000)",
    "bibyears",		"*Range of interesting years for biblio (1950/2000)",
    "mestyp",		"*Choice of measurement display (All)",
    "precoo",		"Maximal precision of coordinates (7)",
    (char *)0
};
static char *initop[] = {
    "mestyp=a",			/* Don't issue the data headers */
    (char *)0
};
static char *initq[] = {
    "/uif set width 800",	/* Be sure errors message complete */
    (char *)0
};
static char blank_string[4] = ""; 	/* For paranoiac compilers...   */

static char usage[] = "\
Usage: rsimbad [-Hhost] [-Sservice] [-Uuser] [-Ppasswd] [-n] [-1] [-pprompt]\n\
       [-m max_data] [parm=val]... [-i input_file|simbad_name]";
static int errors;

static int max_found = 0;
static int optn = 0;
static int verbop = 0;
static int omax = 999999;	/* -m option */
static char *prompt = ">";

/*============================================================================
 *		Error operations
 *============================================================================*/
static void error(msg, arg)
/*++++++++++++++++
.PURPOSE  Message logging
.RETURNS
-----------------*/
  char *msg;	/* IN: Message to print */
  char *arg;	/* IN: Message to print */
{
    fprintf(stderr, "****%s%s\n", msg, arg ? arg : "(nil)");
    errors++;
}

static void help()
/*++++++++++++++++
.PURPOSE  Detailed help
.RETURNS
-----------------*/
{
  static char text[] = "\
   -H specifies the hostname to connect to (default: simbad)\n\
   -S specifies the service  to connect to (default: 1674)\n\
   -U specifies the Simbad username (default prompted)\n\
   -P specifies the Simbad password (default prompted)\n\
   -i specifies the input file (default stdin if no simbad_name argument)\n\
   -m specifies the maxi number of data for an astrotype (e.g. single O.IUE)\n\
   -n gives only numbers of references and measurements (precedence on -m)\n\
   -p specifies the Prompt (default: >), can use C-like controls e.g. \\n\n\
   -1 (or any number) asks to list only the first objects if many are found\n\
   simbad_name represents one simbad name to query, in 1 or more arguments\n\
";
  char **pl;

    fprintf(stderr,"%s\n%s", usage, text);
    fprintf(stderr, "    Available parameters:\n");
    for (pl=params; *pl; pl += 2)
	fprintf(stderr, "\t%-12s %s\n", pl[0], pl[1]);
}

static char *Cstring(str)
/*+++++++
.PURPOSE Interpret a string with C conventions
.RETURNS The transformed string
---------*/
  char *str;    /* IN: string to scan   */
{
  char *s, *p; int i;
    str = strdup(str);
    for (s=p=str; *s; s++, p++) {
	*p = *s;
	if (*s != '\\') continue;
	switch(s[1]) {
	  case 'a': *p = '\a'; s++; continue;
	  case 't': *p = '\t'; s++; continue;
	  case 'n': *p = '\n'; s++; continue;
	  case 'v': *p = '\v'; s++; continue;
	  case 'r': *p = '\r'; s++; continue;
	  case 'f': *p = '\f'; s++; continue;
	}
	if (s[1] < '0') continue;
	if (s[1] > '7') continue;
	i = *++s - '0';
	if ((s[1] >= '0') && (s[1] < '8')) i = (i<<3) | (*++s - '0');
	if ((s[1] >= '0') && (s[1] < '8')) i = (i<<3) | (*++s - '0');
	*p = i;
    }
    *p = 0;
    return(str);
}

/*============================================================================
 *		General Utility Routines
 *============================================================================*/

static int strloc(str, c)
/*+++++++
.PURPOSE Locate the first occurence of character `c'
.RETURNS Index of `c' in str; length of str if `c' not found.
---------*/
  char *str;	/* IN: string to scan	*/
  char c;	/* IN: char to locate	*/
{
  register char *p;

    for (p=str; *p; p++)
  	if (*p == c)	break;
    return(p-str);
}

int strmatch(char *template, char *name)
/*++++++++++++++++
.PURPOSE  Check name against template
.RETURNS  1 (Match) / 0 (mismatch)
.REMARKS  * ? [...] recognized as wild chars
-----------------*/
{
  char *t, *f; int m, not;
    for (t = template, f = name; *t; t++) switch(*t) {
      case '?':			/* Matches one character */
	if (*f) f++;
	else return(0);
	continue;
      case '*':			/* Match 0 or n characters */
	if (! *++t) return(1);	/* The * is the last char of template */
#if 1		/* Try LONGEST possible matching */
	for (m = strlen(f); --m >= 0; ) {
	    if (strmatch(t, f+m))
		return(1);
	}
#else		/* Try SHORTEST possible matching */
	for (; *f; f++) {	/* Look for first matching char */
	    if (isalnum(*t) && (*f != *t)) continue;
	    if (strmatch(t,f))	/* Recursive call */
		return(1);
	}
#endif
	return(0);
      case '[':		/* A set of characters is provided */
	++t; m=not=0;
	if (*t == '^') { not = 1 ; t++ ; }
	if (*t == ']') { /* Closing brace is acce[table	*/
	    t++;
	    m = (*f == ']');
	}
	while ((m==0) && *t) {	/* Check: does *f be in set ? 	*/
	    if (*t == ']') return(0);		/* End of set	*/
	    if ((t[1] == '-') && (t[2] != ']')) {      /* Range	*/
		m = (*f >= t[0]) && (*f <= t[2]);
		t += 3;
	    }
	    else m = (*f == *t), t++;
	}
	m ^= not;	/* V4.9 */
	if (m == 0) return(0);
	t += strloc(t, ']');
	f++;
	continue;
      case '\\':	/* Escaped Char. NO BREAK HERE! */
	t++;
      default:
	if (*t != *f) return(0);
	f++;
	continue;
    }
    return((!*t) && (!*f));	/* Returns 1 if *t AND *f are NUL */
}


#if 0
static char *strdup(str)
/*++++++++++++++++
.PURPOSE  Clone a string
.RETURNS  Newly allocated string
-----------------*/
  char *str ;
{
    char *newstr ;
    newstr = (char *) malloc(strlen(str)+1) ;
    strcpy(newstr,str) ;
    return(newstr) ;
}
#endif

static int alookup(array, str)
/*++++++++++++++++
.PURPOSE  Lookup for a word in an array of word
.RETURNS  Index in array (size if not found)
-----------------*/
  char **array;		/* IN: Array of allowed parameters */
  char *str ;		/* IN: Parameter name (may be followed by =) */
{
  register char **a; char x, *e;

    e = str + strloc(str, '='); x = *e; *e = 0;
    for (a=array; *a; a += 2)
  	if (strcmp(*a, str) == 0 )	break;
    *e = x;
    return(a-array);
}

/*============================================================================
 *		Numeric editions
 *============================================================================*/

static double gf[] = {1., 10., 100., 1.e3,  1.e4,  1.e5,  1.e6,  1.e7,
	1.e8, 1.e9};
static double qf[] = {1., 10.,  60., 600., .36e4, .36e5, .36e6, .36e7,
	.36e8, .36e9 };
			/* Bytes required */
static char gb[] = { 6, 10, 12, 14, 16, 18, 20, 22, 24, 26 };
static char qb[] = { 7, 10, 13, 16, 19, 22, 24, 26, 28, 30 };
/*===========================================================================*/
static int ed(dest, value, nd)
/*++++++++++
.PURPOSE Edit number `value' with `nd' digits
.RETURNS 0 for OK, remainder if number not edited.
-------------*/
  char   *dest; 	/* OUT: pointer to destination string 	*/
  long   value;		/* IN: The number edit	*/
  int    nd;		/* IN: Number of digits	*/
{
  long 	n;
  char 	*p;
  int	i;

    for (n = value, i = nd, p = dest + nd; --i >= 0; n /= 10)
  	*(--p) = '0' + n%10;
    i = n;
    return(i);
}

/*===========================================================================*/
int ed_o1(dest, value, precision, op)
/*++++++++++
.PURPOSE	Edit Single coordinates in decimal degrees
		(longitude or declination)
		The precision (number of decimals) must be between
		1 and 10.
.METHOD		Coordinates are rounded.
.RETURNS	Length of the nul-terminated edited string, -1
		for errors.
-------------*/
  char   *dest; 	/* OUT: pointer to destination string 	*/
  double value[2];	/* IN: The coordinates to edit		*/
  int    precision;	/* IN: Precision for edition		*/
  int 	op;		/* IN: 0 = long, 1 = lat		*/
{
  long 	n;
  char 	*p;
  double f;
  int	nd;	/* Number of Decimals 	*/

	/* Test bad coordinates	or not enough space	*/

    p = dest; *p = 0;

    if (precision <= 0) 	return(0);
    if (precision > 10)	precision = 10;
    nd = precision - 1;

	/* Round the Longitude / Latitude	*/

    if (op == 0) {
   	f = value[0] + .5e0/gf[nd] ;
  	if (f < 0.0e0)		f += 360.0e0;
  	if (f >= 360.0e0)	f -= 360.0e0;
  	n = f;			/* Integer part */
  	ed(p, n, 3);
  	p += 3;
    }
    else {
   	*(p++) = (value[1] >= 0 ? '+' : '-');
  	f = fabs(value[1]) + .5e0/gf[nd] ;
  	n = f;			/* Integer part */
  	ed(p, n, 2);
  	p += 2;
    }

	/* Edit Fractions */

    if (nd > 0) {
   	*(p++) = '.';
  	f = (f - n) * gf[nd];
  	n = f;
  	ed(p, n, nd);
  	p += nd;
    }

    *p = 0;	/* Terminate String */
    return(p - dest);
}

/*===========================================================================*/
int ed_o(dest, value, precision)
/*++++++++++
.PURPOSE	Edit coordinates in decimal degrees.
		The precision (number of decimals) msut be between
		1 and 6.
.METHOD		Coordinates are rounded
.RETURNS	Length of the nul-terminated edited string, -1
		for errors.
-------------*/
  char   *dest; 	/* OUT: pointer to destination string 	*/
  double value[2];	/* IN: The 2 coordinates to edit	*/
  int    precision;	/* IN: Precision for edition		*/
{
  char 	*p;

#ifdef DEBUG
    printf("ed_o: coordinates=(%12.6f,%12.6f)=%d=\n",
	value[0], value[1], precision);
#endif

  p = dest;

	/* Edit the Longitude	*/
  p += ed_o1(p, value, precision, 0);

  *(p++) = ' ';		/* Add a blank	*/

	/* Edit the Latitude	*/
  p += ed_o1(p, value, precision, 1);

  return(p - dest);
}

/*===========================================================================*/
int ed_q1(dest, value, precision, op)
/*++++++++++
.PURPOSE	Edit equatorial coordinates in the astronomical form.
		The precision ranges from 0 to 10, for degrees to
		arc.sec/1000.
.METHOD		Coordinates are rounded
.RETURNS	The length (without the trailing nul) / -1 if error
.REMARKS	Right ascension with at least minutes.
-------------*/
  char 	*dest; 		/* OUT: pointer to destination string 	*/
  double value[2];	/* IN: The 2 coordinates to edit	*/
  int	precision;	/* IN: The precision, 0 to 10		*/
  int   op;		/* IN: 0 = RA, 1 = dec	*/
{
  long 	n;
  char 	*p;
  double f;
  int	nd, i;
  static char sep[3] = {0, '.', ' '};

	/* Test bad coordinates	or not enough space	*/
    p = dest; *p = 0;

    if (precision <= 0) 	return(0);
    if (precision > 10) 	precision = 10;

	/* Round and Edit the Right Ascension	*/
    if (op == 0) {
  	nd = precision; if (nd < 2) nd = 2;
  	f = value[0]/15.0e0 + 0.5e0/qf[nd] ;
  	if (f < 0.0e0)		f += 24.0e0;
  	if (f >= 24.0e0)	f -= 24.0e0;
    }
    else {
  	*(p++) = (value[1] >= 0 ? '+' : '-');
  	nd = precision - 1;
  	f = fabs(value[1]) + .5e0/qf[nd] ;
    }

    n = f;			/* Integer part */
    ed(p, n, 2), p += 2;

	/* Edit mn	*/
    i = MIN(nd, 2);
    if (i == 0)	goto terminate;
    *(p++) = sep[i];
    f = (f-n) * qf[i];
    n = f;
    ed(p, n, i);
    p += i, nd -= i;

	/* Edit seconds	*/
    i = MIN(nd, 2);
    if (i == 0)	goto terminate;
    *(p++) = sep[i];
    f = (f-n) * qf[i];
    n = f;
    ed(p, n, i);
    p += i, nd -= i;

	/* Edit Fractions of seconds	*/
    if (nd == 0)	goto terminate;
    *(p++) = '.';
    f = (f-n) * gf[nd];
    n = f;
    ed(p, n, nd);
    p += nd;

  terminate:
    *p = 0;		/* Terminate String */
    return(p-dest);
}

/*===========================================================================*/
int ed_q(dest, value, precision)
/*++++++++++
.PURPOSE	Edit equatorial coordinates in the astronomical form.
		The precision ranges from 0 to 10, for degrees to
		arc.sec/1000.
.METHOD		Coordinates are rounded
.RETURNS	The length (without the trailing nul) / -1 if error
.REMARKS	The string is filled with * if the coordinates cannot be edited .
-------------*/
  char 	*dest; 		/* OUT: pointer to destination string 	*/
  double value[2];	/* IN: The 2 coordinates to edit	*/
  int	precision;	/* IN: The precision, 0 to 10		*/
{
  char 	*p;

#ifdef DEBUG
    printf("ed_q: coordinates=(%12.6f,%12.6f)=%d=\n",
	value[0], value[1], precision);
#endif

    p = dest;

	/* Edit the Right Ascension	*/
    p += ed_q1(p, value, precision, 0);

    *(p++) = ' ';		/* Add a blank	*/

	/* Edit the Declination	*/
    p += ed_q1(p, value, precision, 1);

    return(p - dest);
}

/*===========================================================================
		Edition with a buffer
 *===========================================================================*/

int ed2_o(buf, size, value, precision)
/*++++++++++
.PURPOSE	Edit Galactic coordinates in the astronomical form.
		The precision ranges from 0 to 10, for degrees to
		fractions of mas
.RETURNS	The length (without the trailing nul) / -1 if error
-------------*/
  char 	*buf; 		/* OUT: pointer to destination string 	*/
  int    size;		/* IN: Szie of buffer			*/
  double value[2];	/* IN: The 2 coordinates to edit	*/
  int	precision;	/* IN: The precision, 0 to 10		*/
{
  char 	*p, *pd;
  int   i, prec;

#ifdef DEBUG
    printf("ed2_o: coordinates=(%12.6f,%12.6f)=%d=\n",
	value[0], value[1], precision);
#endif

    p  = buf; *p = 0;
    if (precision <= 0) return(precision);	/* Not enough space */
    if (precision > ITEMS(gb)) precision = ITEMS(gb);
    for (prec = precision-1; gb[prec] >= size; prec--) ;
    if (prec < 0) return(-1);

	/* Compute where the delta part must start */
    for (i = ITEMS(gb)-1; gb[i] >= size; i-- ) ;
    pd = buf + (size-1) - (gb[i]>>1);

	/* Edit the Longitude		*/
    p += ed_o1(p, value, prec+1, 0);
    	/* Insert blanks to adjust left+right */
    while (p < pd) *(p++) = ' ';

	/* Edit the Declination	*/
    p += ed_o1(p, value, prec+1, 1);
    	/* Insert blanks to adjust left+right */
    for (pd = buf + (size-1); p < pd; p++) *p = ' ';
    *p = 0;

    return(p - buf);
}

/*===========================================================================*/
int ed2_q(buf, size, value, precision)
/*++++++++++
.PURPOSE	Edit equatorial coordinates in the astronomical form.
		The precision ranges from 0 to 10, for degrees to
		fractions of mas
.RETURNS	The length (without the trailing nul) / -1 if error
.REMARKS	The string is filled with * if the coordinates cannot be edited .
-------------*/
  char 	*buf; 		/* OUT: pointer to destination string 	*/
  int    size;		/* IN: Szie of buffer			*/
  double value[2];	/* IN: The 2 coordinates to edit	*/
  int	precision;	/* IN: The precision, 0 to 10		*/
{
  char 	*p, *pd;
  int   i, prec;

#ifdef DEBUG
    printf("ed2_q: coordinates=(%12.6f,%12.6f)=%d=\n",
	value[0], value[1], precision);
#endif

    p  = buf; *p = 0;
    if (precision <= 0) return(precision);	/* Not enough space */
    if (precision >= ITEMS(qb)) precision = ITEMS(qb)-1;
    for (prec = precision; qb[prec] >= size; prec--) ;
    if (prec < 0) return(-1);

	/* Compute where the delta part must start */
    for (i = ITEMS(qb)-1; qb[i] >= size; i-- ) ;
    pd = buf + (size-1) - (qb[i]>>1);

	/* Edit the Right Ascension	*/
    p += ed_q1(p, value, prec, 0);
    	/* Insert blanks to adjust left+right */
    while (p < pd) *(p++) = ' ';

	/* Edit the Declination	*/
    p += ed_q1(p, value, prec, 1);
    	/* Insert blanks to adjust left+right */
    for (pd = buf + (size-1); p < pd; p++) *p = ' ';
    *p = 0;

    return(p - buf);
}

/*===========================================================================*/
int print_q(data)
/*++++++++++
.PURPOSE Print out the coordinates in sexagesimal
.RETURNS Number of bytes written
.REMARKS Input string assumed to be the "J" astrotype
-------------*/
  char *data;		/* IN: Rssult of a SIMBAD query	*/
{
  char *s; double o[2]; int prec, n;
  char acoo[28];
  double atof();
  static int precoo = 0;

    if (! precoo) {
	if ((s=getenv("precoo"))) precoo = atoi(s);
	else precoo = 7;
    }

    for (s=data; isspace(*s); s++) ;
    o[0] = atof(s);
    while (isgraph(*s)) s++;
    o[1] = atof(s);
    s += strloc(s, '(');
    if (*s) prec = atoi(s+1);
    else prec = 5;		/* 1" */
    if (prec > precoo) prec = precoo ;
    n = ed2_q(acoo, sizeof(acoo), o, prec);
    if (n > 0) printf("%s", acoo);
    return(n);
}

/*============================================================================
 *		Start up SIMBAD
 *============================================================================*/

static int init()
/*++++++++++++++++
.PURPOSE  Check hostname, etc, and open the Simbad session
.RETURNS
.REMARKS
-----------------*/
{
#if 0	/* Version with password */
  char buf[256];
  char *p;
#endif
  int hs;	/* Handle for Simbad */
  char **a;

    if (!hostname) hostname = HOSTNAME;
    if (!service)  service = SERVICE;
    if (!username) {
	username = getenv("USER");
#if 0
    	fprintf(stderr,"simbad client userid [%s]: ", username) ;
	if (!fgets(buf, sizeof(buf), stdin)) exit(1);
	for (p = buf+strlen(buf)-1; (p>=buf) && (isspace(*p)); p--) *p = 0;
	if (*buf) username = buf;
#else
	if (!username) username = "v1332viz";
#endif
    }
    if (!password) {
#if 0	/* No password... */
	password = getpass("simbad client passwd: ") ;
#else
	password = "sim4Viz";
#endif
    }
    hs = simbad_connect(hostname, service, username, password);
    if (hs < 0) {
	error("Connecting Simbad: ", simbad_error(0));
	exit(1);
    }

	/* Sent the initialisations */
    for (a=initop; *a; a++) {
	if (simbad_query(hs, blank_string, *a) < 0) {
	    error("Initialize Simbad Queries", simbad_error(hs));
	    exit(1);
	}
    }
    for (a=initq; *a; a++) {
	if (simbad_query(hs, *a, "")< 0)
	  error(*a, simbad_error(hs));
    }

    return(hs);
}

static int print_data(int hs, char *asked_type, char *types)
/*++++++++++++++++
.PURPOSE  Print the selected astrotype
.RETURNS  Number of lines written out
.REMARKS  types may be NULL, in which case it's retrieved.
-----------------*/
{
  char *data;
  int ndat, hash, lines=0;

    if ((hash = (*asked_type == '#'))) 	/* Just number of values */
	asked_type++;
    if ((!types) || (asked_type == types))
	goto find_and_print;

    /* Are there wild cards ? If yes, a recursive call will do the job... */
    for (data=asked_type; *data; data++) {
	if ((*data == '*') || (*data == '?') || (*data == '['))
	    break;
    }
    if (*data) {	/* There are wild cards! */
      int i; char typ[16];
        if (!types) types = simbad_telldata(hs);
        typ[0] = '#'; 		/* Will be replaced if # not asked */
	for (data=types; isspace(*data); data++) ;
	while (*data) {
	    for (i=0; isgraph(data[i]); i++) ;
	    if (i>=sizeof(typ)-2) {
		fprintf(stderr,
			"#***rsimbad: bad astrotype returned: %s\n", data);
		i = sizeof(typ)-2;
	    }
	    memcpy(typ+hash, data, i); typ[hash+i] = 0;
	    if (strmatch(asked_type, typ))
		lines += print_data(hs, typ, typ);
	    data += i; while (isspace(*data)) data++;
	}
	return(lines);
    }

    /* No wild char -- get the result and print it */
  find_and_print:
    ndat = simbad_findata(hs, asked_type, blank_string);
    if (hash)
	printf("#%s %d\n", asked_type, ndat);
    else {
	if (ndat > omax) ndat = omax;
	while (--ndat >= 0) {
            data = simbad_getdata(hs, 0);
            if (!data) break;
	    if (isspace(*data)) printf("%%%s%s", asked_type, data);
	    else               printf("%%%s %s", asked_type, data);
            if ((asked_type[0] == 'J') && (asked_type[1] == '\0')) {
		/* Coordinates, add the sexagesimal */
    	        printf(" = ");
    	        print_q(data);
	    }
            putchar('\n');
        }
    }
    return(lines);
}

/*===========================================================================*/

#define aVal(p)		if (p[2]) p += 2; else --ac, p = *++a

int main(argc, argv) int argc; char **argv;
/*++++++++++++++++
.PURPOSE
.RETURNS  to OS
.REMARKS  Options -Sservice -Uuser -Ppasswd
-----------------*/
{
  time_t tim; char **a; int ac;
  char buf[1024];
  char nastrotypes[1024];
  FILE *infile; char *input_name;
  char *parm, *p, *q, *query, *abuf, *tofree, x;
  int  hs = -1;		/* Handle for Simbad */
  int found, nobj, i;
  int goon = 1;

	/* Check the Arguments */
    query = input_name = (char *)0;
    infile = stdin;
    abuf   = buf;
    for (ac=argc, a=argv; --ac > 0; ) {
	parm = *++a;
	if (*parm == '-') switch(parm[1]) {
	    case 'h':	/* Help */
			help(); exit(0);
	    case 'p':	aVal(parm); prompt = Cstring(parm); continue;
	    case 'S':	aVal(parm); service  = parm; continue;
	    case 'H':	aVal(parm); hostname = parm; continue;
	    case 'U':	aVal(parm); username = parm; continue;
	    case 'P':	aVal(parm); password = strdup(parm);
			while (*parm) *(parm++) = 0;
			continue;
	    case 'm':	aVal(parm); omax = atoi(parm);
			if (omax<=0) omax=999999;
			continue;
	    case '#':
	    case 'n':	optn = 1; continue;
	    case 'v':   if (strncmp(parm, "-vers", 5) == 0) {
			    printf("rsimbad (CDS) Version %s\n", VERSION);
			    exit(0);
			}
			verbop = 1;
			continue;
	    case 'q':	/* Kept for compatibility version 1.4 */
		if (query) fprintf(stderr,
		    "++++rsimbad: a single -q option allowed, '%s' discarded\n",
		    query);
	        query = *++a; --ac;
		continue;
	    case 'i':	/* Input file */
	        if (infile != stdin)
		    error("a single option allowed", parm);
		aVal(parm);
	        input_name = parm;
	        infile = fopen(input_name, "r");
	        if (!infile) {
	            fprintf(stderr, "****rsimbad: ");
	            perror(input_name);
	            exit(1);
	        }
	        continue;
	    /*
	    case '1':	max_found = 1;     continue;
	    */
	    default:
	      if (isdigit(parm[1])) { max_found = atoi(parm+1); continue; }
		error("Unknown option: ", parm);
		continue;
	}
	if (parm[strloc(parm, '=')]) {	/* Something like  param=value */
	    i = alookup(params, parm);
	    if (params[i]) { 	/* Valid parameter. Keep & Mail to Simbad */
		if (*params[i+1] == ',') 	/* ... Translate commas   */
		    for (p=parm; *p; p++) if (*p == ',') *p = ' ';
		if (hs < 0) hs = init();
		putenv(parm);
		if (*params[i+1] == '*') {
    		    if (simbad_query(hs, blank_string, parm) < 0)
			error("Simbad Option", simbad_error(hs));
		}
		continue;
	    }
	    error("Unknown parameter: ", parm);
	    continue;
	}
	/* It's an object name ! */
	argc = a-argv;			/* Count of 'real' arguments */
	for (i=found=0; i<ac; i++) found += 1+strlen(a[i]);
	query = p = malloc(found+1);	/* Copy the object name...   */
	strcpy(p, *a);
	for (i=1; i<ac; i++) {
	    p += strlen(p);
	    *(p++) = ' ';
	    strcpy(p, a[i]);
	}
	break;
    }
    if(errors) exit(1);

	/* Print the Prologue (except when -help option) */
    time(&tim);
    printf("%%%%<SIMBAD> (%s)\t\t\t\t%s%%%%Args:",
	    hostname ? hostname : "default", ctime(&tim));
    fflush(stdout);
    for (ac=argc, a=argv; --ac > 0; ) {
	parm = *++a;
	i = strloc(parm, '=');
	if (parm[i]) {	/* Something like  param=value */
	    x=parm[i]; parm[i]=0;
	    printf(" %s='%s'", parm, parm+i+1);
	    parm[i] = x;
	} else printf(" %s", parm);
    }
    printf("\n");

	/* Connect to simbad */
    if (hs < 0) hs = init();
    printf("\n");

	/* Each line contains a question; separate options ; and search */

    while(goon) {
	tofree = (char *)0;
	if (query) { 
            if (verbop) fprintf(stderr, 
		    "#...rsimbad_input: query=%s\n", query);
	    abuf = query; 
	    goon = 0; 
	    printf("%s", prompt);
	}
	else if (isatty(fileno(infile))) {
#if HAS_READLINE
	    if (!history_file) {
		using_history(); 
		stifle_history(1000);
		history_file = getenv("HISTORY_FILE");
		if (!history_file) {
		    p = getenv("HOME");
		    i = p ? strlen(p)+1 : 0;
		    history_file = malloc(i+1+sizeof(HISTORY_FILE));
		    if (p) { 
			strcpy(history_file, p); 
			strcat(history_file, "/");
		    }
		    else *history_file = 0;
		    strcat(history_file, HISTORY_FILE);
		}
		if (read_history(history_file)) 
		    write_history(history_file);
	    }
            if (verbop) printf("#...rsimbad_input: isatty+readline...\n");
	    abuf = tofree = readline(prompt);
	    printf("%s", prompt);
#else
            if (verbop) printf("#...rsimbad_input: isatty...\n");
	    printf("%s", prompt); fflush(stdout);
	    abuf = fgets(buf, sizeof(buf), infile);
#endif
	}
	else {
            if (verbop) printf("#...rsimbad_input:  no_tty...\n");
	    printf("%s", prompt); fflush(stdout);
	    abuf = fgets(buf, sizeof(buf), infile);
	}
	if (!abuf) { goon = 0; continue; }
	for (p = abuf+strlen(abuf)-1; (p>=abuf) && (isspace(*p)); p--) 
	    *p = 0;
	printf("%s", abuf);
	/* Blank lines, lines starting by # or . or : are a comment */
	if ((!abuf[0]) || strchr("#.:", abuf[0])) {
	    puts("");
	    continue;
	}
	printf("\t!");

		/* Send the Question to Simbad */
	found = simbad_query(hs, abuf, (char *)0);
#if HAS_READLINE
	if (isatty(fileno(infile))) add_history(abuf);
#endif
	if (tofree) free(tofree);
	if (found < 0) {
	    parm =  simbad_error(hs);
	    printf("*** %s\n\n", parm);
	    continue;
	}

		/* Write the Number of Objects found */
	printf("=%d=\n", found);
	if (found == 0) { printf("\n"); continue; }

		/* Loop on the objects found */
	if (max_found > 0) found = found < max_found ? found : max_found ;
	for (nobj=1; nobj <= found; nobj++) {
	    printf("=%d/%d\t!", nobj, found);
	    if (simbad_retrieve(hs, nobj) < 0) {
		parm = simbad_error(hs);
	    	printf(" #*** %s\n\n", parm);
		continue;
	    }

	    q = simbad_telldata(hs);
	    printf(" %s\n", q);

		/* Display data asked for */
	    if (optn) {		/* Ask to give only NUMBER of data */
		for (p=nastrotypes;
		    *q && (p < &nastrotypes[sizeof(nastrotypes)]-1);
		    q += i, p += i) {
		    i = strloc(q, ' ');
		    if (i == 0) { *p = ' '; i=1; continue; }
		    if ((i == 1) && (q[0] == 'B' || q[0] == 'R'))
			*(p++) = '#';		/* Reference */
		    else if ((i>2) && (q[0] == 'O') && (q[1] == '.'))
			*(p++) = '#';		/* Measurement */
		    strcpy(p, q);
		}
		*p = 0;
		q = nastrotypes;
	    }
	    p=getenv("astrotypes");
	    if (!p) p = q;
	    if (!*p) p = q;
	    while (*p) {
		i = strloc(p, ' '); x = p[i]; p[i] = 0;
#if 0
		if (*p == '#') 		/* Just number of items */
		    printf("%s %d\n", p, simbad_findata(hs, p+1, blank_string));
		else for (ndat = simbad_findata(hs, p, blank_string);
		  --ndat >= 0; ) {
		    data = simbad_getdata(hs, 0);
		    if (!data) break;
		    printf("%%%s", p) ;
		    if (isspace(*data)) ; else putchar(' ') ;
		    printf("%s", data);
		    if ((p[0] == 'J') && (p[1] == '\0')) {     /* Coordinates */
			printf(" = ");
			print_q(data);
		    }
		    putchar('\n');
		}
		p[i] = x;
		while (isgraph(*p)) p++;
		while (isspace(*p)) p++;
#else
		print_data(hs, p, q);
		p[i] = x;
		for (p += i; isspace(*p); p++) ;
#endif
	    }
	    printf ("\n");
	}
    }
    time(&tim);
    printf("</SIMBAD>\t\t\t\t\t%s", ctime(&tim));
#if HAS_READLINE
    if (history_file) write_history(history_file);
    /*history_truncate_file((char *)0, 1000); */
#endif
    simbad_disconnect(hs);
    exit(0) ;
}
